"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DatabaseChanges = void 0;
const DatabaseConnectionState_1 = require("./DatabaseConnectionState");
const ChangesObservable_1 = require("./ChangesObservable");
const Exceptions_1 = require("../../Exceptions");
const semaphore = require("semaphore");
const WebSocket = require("ws");
const StringUtil_1 = require("../../Utility/StringUtil");
const events_1 = require("events");
const PromiseUtil = require("../../Utility/PromiseUtil");
const SemaphoreUtil_1 = require("../../Utility/SemaphoreUtil");
const Certificate_1 = require("../../Auth/Certificate");
const ObjectUtil_1 = require("../../Utility/ObjectUtil");
const ServerNode_1 = require("../../Http/ServerNode");
const UpdateTopologyParameters_1 = require("../../Http/UpdateTopologyParameters");
const TypeUtil_1 = require("../../Utility/TypeUtil");
const AggressiveCacheChange_1 = require("./AggressiveCacheChange");
class DatabaseChanges {
    constructor(requestExecutor, databaseName, onDispose, nodeTag) {
        this._emitter = new events_1.EventEmitter();
        this._commandId = 0;
        this._semaphore = semaphore();
        this._isCanceled = false;
        this._confirmations = new Map();
        this._counters = new Map();
        this._immediateConnection = 0;
        this._requestExecutor = requestExecutor;
        this._conventions = requestExecutor.conventions;
        this._database = databaseName;
        this._tcs = PromiseUtil.defer();
        this._onDispose = onDispose;
        this._onConnectionStatusChangedWrapped = () => this._onConnectionStatusChanged();
        this._emitter.on("connectionStatus", this._onConnectionStatusChangedWrapped);
        this._task = this._doWork(nodeTag);
    }
    static createClientWebSocket(requestExecutor, url) {
        const authOptions = requestExecutor.getAuthOptions();
        let options = undefined;
        if (authOptions) {
            const certificate = Certificate_1.Certificate.createFromOptions(authOptions);
            options = certificate.toWebSocketOptions();
        }
        return new WebSocket(url, options);
    }
    async _onConnectionStatusChanged() {
        const acquiredSemContext = (0, SemaphoreUtil_1.acquireSemaphore)(this._semaphore);
        try {
            await acquiredSemContext.promise;
            if (this.connected) {
                this._tcs.resolve(this);
                return;
            }
            if (this._tcs.promise.isFulfilled()) {
                this._tcs = PromiseUtil.defer();
            }
        }
        finally {
            acquiredSemContext.dispose();
        }
    }
    get connected() {
        return this._client && this._client.readyState === WebSocket.OPEN;
    }
    on(eventName, handler) {
        this._emitter.addListener(eventName, handler);
        return this;
    }
    off(eventName, handler) {
        this._emitter.removeListener(eventName, handler);
        return this;
    }
    ensureConnectedNow() {
        return Promise.resolve(this._tcs.promise);
    }
    forIndex(indexName) {
        if (StringUtil_1.StringUtil.isNullOrWhitespace(indexName)) {
            (0, Exceptions_1.throwError)("InvalidArgumentException", "IndexName cannot be null or whitespace.");
        }
        const counter = this._getOrAddConnectionState("indexes/" + indexName, "watch-index", "unwatch-index", indexName);
        return new ChangesObservable_1.ChangesObservable("Index", counter, notification => notification.name
            && notification.name.toLocaleLowerCase() === indexName.toLocaleLowerCase());
    }
    get lastConnectionStateException() {
        for (const counter of Array.from(this._counters.values())) {
            if (counter.lastError) {
                return counter.lastError;
            }
        }
        return null;
    }
    forDocument(docId) {
        if (StringUtil_1.StringUtil.isNullOrWhitespace(docId)) {
            (0, Exceptions_1.throwError)("InvalidArgumentException", "DocumentId cannot be null or whitespace.");
        }
        const counter = this._getOrAddConnectionState("docs/" + docId, "watch-doc", "unwatch-doc", docId);
        return new ChangesObservable_1.ChangesObservable("Document", counter, notification => notification.id && notification.id.toLocaleLowerCase() === docId.toLocaleLowerCase());
    }
    forAllDocuments() {
        const counter = this._getOrAddConnectionState("all-docs", "watch-docs", "unwatch-docs", null);
        return new ChangesObservable_1.ChangesObservable("Document", counter, () => true);
    }
    forOperationId(operationId) {
        const counter = this._getOrAddConnectionState("operations/" + operationId, "watch-operation", "unwatch-operation", operationId.toString());
        return new ChangesObservable_1.ChangesObservable("Operation", counter, notification => notification.operationId === operationId);
    }
    forAllOperations() {
        const counter = this._getOrAddConnectionState("all-operations", "watch-operations", "unwatch-operations", null);
        return new ChangesObservable_1.ChangesObservable("Operation", counter, () => true);
    }
    forAllIndexes() {
        const counter = this._getOrAddConnectionState("all-indexes", "watch-indexes", "unwatch-indexes", null);
        return new ChangesObservable_1.ChangesObservable("Index", counter, () => true);
    }
    forDocumentsStartingWith(docIdPrefix) {
        if (StringUtil_1.StringUtil.isNullOrWhitespace(docIdPrefix)) {
            (0, Exceptions_1.throwError)("InvalidArgumentException", "DocumentId cannot be null or whitespace.");
        }
        const counter = this._getOrAddConnectionState("prefixes/" + docIdPrefix, "watch-prefix", "unwatch-prefix", docIdPrefix);
        return new ChangesObservable_1.ChangesObservable("Document", counter, notification => notification.id
            && notification.id.toLocaleLowerCase().startsWith(docIdPrefix.toLocaleLowerCase()));
    }
    forDocumentsInCollection(collectionNameOrDescriptor) {
        const collectionName = typeof collectionNameOrDescriptor !== "string"
            ? this._conventions.getCollectionNameForType(collectionNameOrDescriptor)
            : collectionNameOrDescriptor;
        if (!collectionName) {
            (0, Exceptions_1.throwError)("InvalidArgumentException", "CollectionName cannot be null");
        }
        const counter = this._getOrAddConnectionState("collections/" + collectionName, "watch-collection", "unwatch-collection", collectionName);
        return new ChangesObservable_1.ChangesObservable("Document", counter, notification => notification.collectionName
            && collectionName.toLocaleLowerCase() === notification.collectionName.toLocaleLowerCase());
    }
    dispose() {
        for (const confirmation of this._confirmations.values()) {
            confirmation.reject();
        }
        this._isCanceled = true;
        if (this._client) {
            this._client.close();
        }
        for (const value of this._counters.values()) {
            value.dispose();
        }
        this._counters.clear();
        this._emitter.emit("connectionStatus");
        this._emitter.removeListener("connectionStatus", this._onConnectionStatusChangedWrapped);
        if (this._onDispose) {
            this._onDispose();
        }
    }
    _getOrAddConnectionState(name, watchCommand, unwatchCommand, value, values = null) {
        let newValue = false;
        let counter;
        if (!this._counters.has(name)) {
            const connectionState = new DatabaseConnectionState_1.DatabaseConnectionState(() => this._send(watchCommand, value, values), async () => {
                try {
                    if (this.connected) {
                        await this._send(unwatchCommand, value, values);
                    }
                }
                catch (e) {
                }
                const state = this._counters.get(name);
                this._counters.delete(name);
                state.dispose();
            });
            this._counters.set(name, connectionState);
            counter = connectionState;
            newValue = true;
        }
        else {
            counter = this._counters.get(name);
        }
        if (newValue && this._immediateConnection) {
            counter.set(counter.onConnect());
        }
        return counter;
    }
    _send(command, value, values) {
        return new Promise((async (resolve, reject) => {
            let currentCommandId;
            const acquiredSemContext = (0, SemaphoreUtil_1.acquireSemaphore)(this._semaphore, {
                timeout: 15000,
                contextName: "DatabaseChanges._send()"
            });
            try {
                await acquiredSemContext.promise;
                currentCommandId = ++this._commandId;
                const payload = {
                    CommandId: currentCommandId,
                    Command: command,
                    Param: value
                };
                if (values && values.length) {
                    payload["Params"] = values;
                }
                this._confirmations.set(currentCommandId, { resolve, reject });
                const payloadAsString = JSON.stringify(payload, null, 0);
                this._client.send(payloadAsString);
            }
            catch (err) {
                if (!this._isCanceled) {
                    throw err;
                }
            }
            finally {
                if (acquiredSemContext) {
                    acquiredSemContext.dispose();
                }
            }
        }));
    }
    async _doWork(nodeTag) {
        let preferredNode;
        try {
            preferredNode = nodeTag || this._requestExecutor.conventions.disableTopologyUpdates
                ? await this._requestExecutor.getRequestedNode(nodeTag)
                : await this._requestExecutor.getPreferredNode();
            this._nodeIndex = preferredNode.currentIndex;
            this._serverNode = preferredNode.currentNode;
        }
        catch (e) {
            this._emitter.emit("connectionStatus");
            this._notifyAboutError(e);
            this._tcs.reject(e);
            return;
        }
        this._doWorkInternal();
    }
    _doWorkInternal() {
        if (this._isCanceled) {
            return;
        }
        let wasConnected = false;
        if (!this.connected) {
            const urlString = this._serverNode.url + "/databases/" + this._database + "/changes";
            const url = StringUtil_1.StringUtil.toWebSocketPath(urlString);
            this._client = DatabaseChanges.createClientWebSocket(this._requestExecutor, url);
            this._client.on("open", async () => {
                wasConnected = true;
                this._immediateConnection = 1;
                for (const counter of this._counters.values()) {
                    counter.set(counter.onConnect());
                }
                this._emitter.emit("connectionStatus");
            });
            this._client.on("error", async (e) => {
                if (wasConnected) {
                    this._emitter.emit("connectionStatus");
                }
                wasConnected = false;
                try {
                    this._serverNode = await this._requestExecutor.handleServerNotResponsive(this._url, this._serverNode, this._nodeIndex, e);
                }
                catch (ee) {
                    if (ee.name === "DatabaseDoesNotExistException") {
                        e = ee;
                        throw ee;
                    }
                    else {
                    }
                }
                this._notifyAboutError(e);
            });
            this._client.on("close", () => {
                if (this._reconnectClient()) {
                    setTimeout(() => this._doWorkInternal(), 1000);
                }
                for (const confirm of this._confirmations.values()) {
                    confirm.reject();
                }
                this._confirmations.clear();
            });
            this._client.on("message", async (data) => {
                await this._processChanges(data);
            });
        }
    }
    _reconnectClient() {
        if (this._isCanceled) {
            return false;
        }
        this._client.close();
        this._immediateConnection = 0;
        return true;
    }
    async _processChanges(data) {
        if (this._isCanceled) {
            return;
        }
        const payloadParsed = JSON.parse(data);
        try {
            const messages = Array.isArray(payloadParsed) ? payloadParsed : [payloadParsed];
            for (const message of messages) {
                const type = message.Type;
                if (message.TopologyChange) {
                    const state = this._getOrAddConnectionState("Topology", "watch-topology-change", "", "");
                    state.addOnError(TypeUtil_1.TypeUtil.NOOP);
                    const updateParameters = new UpdateTopologyParameters_1.UpdateTopologyParameters(this._serverNode);
                    updateParameters.timeoutInMs = 0;
                    updateParameters.forceUpdate = true;
                    updateParameters.debugTag = "watch-topology-change";
                    this._requestExecutor.updateTopology(updateParameters);
                    continue;
                }
                if (!type) {
                    continue;
                }
                switch (type) {
                    case "Error": {
                        const exceptionAsString = message.Exception;
                        this._notifyAboutError(exceptionAsString);
                        break;
                    }
                    case "Confirm": {
                        const commandId = message.CommandId;
                        const confirmationResolver = this._confirmations.get(commandId);
                        if (confirmationResolver) {
                            confirmationResolver.resolve();
                            this._confirmations.delete(commandId);
                        }
                        break;
                    }
                    default: {
                        const value = message.Value;
                        let transformedValue = ObjectUtil_1.ObjectUtil.transformObjectKeys(value, { defaultTransform: "camel" });
                        if (type === "TimeSeriesChange") {
                            const dateUtil = this._conventions.dateUtil;
                            const timeSeriesValue = transformedValue;
                            const overrides = {
                                from: dateUtil.parse(timeSeriesValue.from),
                                to: dateUtil.parse(timeSeriesValue.to)
                            };
                            transformedValue = Object.assign(transformedValue, overrides);
                        }
                        this._notifySubscribers(type, transformedValue);
                        break;
                    }
                }
            }
        }
        catch (err) {
            this._notifyAboutError(err);
            (0, Exceptions_1.throwError)("ChangeProcessingException", "There was an error during notification processing.", err);
        }
    }
    _notifySubscribers(type, value) {
        switch (type) {
            case "AggressiveCacheChange":
                this._counters.forEach(state => state.send("AggressiveCache", AggressiveCacheChange_1.AggressiveCacheChange.INSTANCE));
                break;
            case "DocumentChange":
                this._counters.forEach(state => state.send("Document", value));
                break;
            case "CounterChange":
                this._counters.forEach(state => state.send("Counter", value));
                break;
            case "TimeSeriesChange":
                this._counters.forEach(state => state.send("TimeSeries", value));
                break;
            case "IndexChange":
                this._counters.forEach(state => state.send("Index", value));
                break;
            case "OperationStatusChange":
                this._counters.forEach(state => state.send("Operation", value));
                break;
            case "TopologyChange": {
                const topologyChange = value;
                const requestExecutor = this._requestExecutor;
                if (requestExecutor) {
                    const node = new ServerNode_1.ServerNode({
                        url: topologyChange.url,
                        database: topologyChange.database
                    });
                    const updateParameters = new UpdateTopologyParameters_1.UpdateTopologyParameters(node);
                    updateParameters.timeoutInMs = 0;
                    updateParameters.forceUpdate = true;
                    updateParameters.debugTag = "topology-change-notification";
                    requestExecutor.updateTopology(updateParameters);
                }
                break;
            }
            default:
                (0, Exceptions_1.throwError)("NotSupportedException");
        }
    }
    _notifyAboutError(e) {
        if (this._isCanceled) {
            return;
        }
        this._emitter.emit("error", e);
        for (const state of this._counters.values()) {
            state.error(e);
        }
    }
    forAllCounters() {
        const counter = this._getOrAddConnectionState("all-counters", "watch-counters", "unwatch-counters", null);
        const taskedObservable = new ChangesObservable_1.ChangesObservable("Counter", counter, notification => true);
        return taskedObservable;
    }
    forCounter(counterName) {
        if (StringUtil_1.StringUtil.isNullOrWhitespace(counterName)) {
            (0, Exceptions_1.throwError)("InvalidArgumentException", "CounterName cannot be null or whitespace.");
        }
        const counter = this._getOrAddConnectionState("counter/" + counterName, "watch-counter", "unwatch-counter", counterName);
        const taskedObservable = new ChangesObservable_1.ChangesObservable("Counter", counter, notification => StringUtil_1.StringUtil.equalsIgnoreCase(counterName, notification.name));
        return taskedObservable;
    }
    forCounterOfDocument(documentId, counterName) {
        if (StringUtil_1.StringUtil.isNullOrWhitespace(documentId)) {
            (0, Exceptions_1.throwError)("InvalidArgumentException", "DocumentId cannot be null or whitespace.");
        }
        if (StringUtil_1.StringUtil.isNullOrWhitespace(counterName)) {
            (0, Exceptions_1.throwError)("InvalidArgumentException", "CounterName cannot be null or whitespace.");
        }
        const counter = this._getOrAddConnectionState("document/" + documentId + "/counter/" + counterName, "watch-document-counter", "unwatch-document-counter", null, [documentId, counterName]);
        const taskedObservable = new ChangesObservable_1.ChangesObservable("Counter", counter, notification => StringUtil_1.StringUtil.equalsIgnoreCase(documentId, notification.documentId)
            && StringUtil_1.StringUtil.equalsIgnoreCase(counterName, notification.name));
        return taskedObservable;
    }
    forCountersOfDocument(documentId) {
        if (StringUtil_1.StringUtil.isNullOrWhitespace(documentId)) {
            (0, Exceptions_1.throwError)("InvalidArgumentException", "DocumentId cannot be null or whitespace.");
        }
        const counter = this._getOrAddConnectionState("document/" + documentId + "/counter", "watch-document-counters", "unwatch-document-counters", documentId);
        const taskedObservable = new ChangesObservable_1.ChangesObservable("Counter", counter, notification => StringUtil_1.StringUtil.equalsIgnoreCase(documentId, notification.documentId));
        return taskedObservable;
    }
    forAllTimeSeries() {
        const counter = this._getOrAddConnectionState("all-timeseries", "watch-all-timeseries", "unwatch-all-timeseries", null);
        const taskedObservable = new ChangesObservable_1.ChangesObservable("TimeSeries", counter, () => true);
        return taskedObservable;
    }
    forTimeSeries(timeSeriesName) {
        if (StringUtil_1.StringUtil.isNullOrWhitespace(timeSeriesName)) {
            (0, Exceptions_1.throwError)("InvalidArgumentException", "TimeSeriesName cannot be null or whitespace.");
        }
        const counter = this._getOrAddConnectionState("timeseries/" + timeSeriesName, "watch-timeseries", "unwatch-timeseries", timeSeriesName);
        const taskedObservable = new ChangesObservable_1.ChangesObservable("TimeSeries", counter, notification => StringUtil_1.StringUtil.equalsIgnoreCase(timeSeriesName, notification.name));
        return taskedObservable;
    }
    forTimeSeriesOfDocument(documentId, timeSeriesName) {
        if (timeSeriesName) {
            return this._forTimeSeriesOfDocumentWithNameInternal(documentId, timeSeriesName);
        }
        else {
            return this._forTimeSeriesOfDocumentInternal(documentId);
        }
    }
    _forTimeSeriesOfDocumentInternal(documentId) {
        if (StringUtil_1.StringUtil.isNullOrWhitespace(documentId)) {
            (0, Exceptions_1.throwError)("InvalidArgumentException", "DocumentId cannot be null or whitespace.");
        }
        const counter = this._getOrAddConnectionState("document/" + documentId + "/timeseries", "watch-all-document-timeseries", "unwatch-all-document-timeseries", documentId);
        const taskedObservable = new ChangesObservable_1.ChangesObservable("TimeSeries", counter, notification => StringUtil_1.StringUtil.equalsIgnoreCase(documentId, notification.documentId));
        return taskedObservable;
    }
    _forTimeSeriesOfDocumentWithNameInternal(documentId, timeSeriesName) {
        if (StringUtil_1.StringUtil.isNullOrWhitespace(documentId)) {
            (0, Exceptions_1.throwError)("InvalidArgumentException", "DocumentId cannot be null or whitespace.");
        }
        if (StringUtil_1.StringUtil.isNullOrWhitespace(timeSeriesName)) {
            (0, Exceptions_1.throwError)("InvalidArgumentException", "TimeSeriesName cannot be null or whitespace.");
        }
        const counter = this._getOrAddConnectionState("document/" + documentId + "/timeseries/" + timeSeriesName, "watch-document-timeseries", "unwatch-document-timeseries", null, [documentId, timeSeriesName]);
        const taskedObservable = new ChangesObservable_1.ChangesObservable("TimeSeries", counter, notification => StringUtil_1.StringUtil.equalsIgnoreCase(timeSeriesName, notification.name)
            && StringUtil_1.StringUtil.equalsIgnoreCase(documentId, notification.documentId));
        return taskedObservable;
    }
}
exports.DatabaseChanges = DatabaseChanges;
//# sourceMappingURL=DatabaseChanges.js.map